@@ -90,6 +90,7 @@ static void write_margin(struct generator * g) {
 }
 
 static void write_comment(struct generator * g, struct node * p) {
+    if (!g->options->comments) return;
     write_margin(g);
     write_string(g, "# ");
     write_comment_content(g, p);
@@ -331,8 +332,8 @@ static void generate_or(struct generator * g, struct node * p) {
         exit(1);
     }
     while (p->right != 0) {
-        g->failure_label = new_label(g);
-        int label = g->failure_label;
+        int label = new_label(g);
+        g->failure_label = label;
         wsetlab_begin(g);
         generate(g, p);
         if (!g->unreachable) {
@@ -374,16 +375,15 @@ static void generate_not(struct generator * g, struct node * p) {
 
     int a0 = g->failure_label;
     struct str * a1 = str_copy(g->failure_str);
+    int label = new_label(g);
+    g->failure_label = label;
+    str_clear(g->failure_str);
 
     write_comment(g, p);
     if (keep_c) {
         write_savecursor(g, p, savevar);
     }
 
-    g->failure_label = new_label(g);
-    int label = g->failure_label;
-    str_clear(g->failure_str);
-
     wsetlab_begin(g);
 
     generate(g, p->left);
@@ -406,14 +406,15 @@ static void generate_try(struct generator * g, struct node * p) {
 
     struct str * savevar = vars_newname(g);
     int keep_c = K_needed(g, p->left);
+    int label = new_label(g);
+    g->failure_label = label;
+    str_clear(g->failure_str);
 
     write_comment(g, p);
-    if (keep_c) write_savecursor(g, p, savevar);
-
-    g->failure_label = new_label(g);
-    int label = g->failure_label;
-
-    if (keep_c) restore_string(p, g->failure_str, savevar);
+    if (keep_c) {
+        write_savecursor(g, p, savevar);
+        restore_string(p, g->failure_str, savevar);
+    }
 
     wsetlab_begin(g);
     generate(g, p->left);
@@ -480,8 +481,8 @@ static void generate_do(struct generator * g, struct node * p) {
         g->V[0] = p->left->name;
         w(g, "~M~V0()~N");
     } else {
-        g->failure_label = new_label(g);
-        int label = g->failure_label;
+        int label = new_label(g);
+        g->failure_label = label;
         str_clear(g->failure_str);
 
         wsetlab_begin(g);
@@ -497,9 +498,7 @@ static void generate_do(struct generator * g, struct node * p) {
 static void generate_GO_grouping(struct generator * g, struct node * p, int is_goto, int complement) {
 
     struct grouping * q = p->name->grouping;
-    w(g, (is_goto ? "~M# goto " : "~M# gopast "));
-    write_comment_content(g, p);
-    write_newline(g);
+    write_comment(g, p);
     g->S[0] = p->mode == m_forward ? "" : "_b";
     g->S[1] = complement ? "in" : "out";
     g->S[2] = g->options->encoding == ENC_UTF8 ? "_U" : "";
@@ -516,6 +515,15 @@ static void generate_GO_grouping(struct generator * g, struct node * p, int is_g
 }
 
 static void generate_GO(struct generator * g, struct node * p, int style) {
+    int end_unreachable;
+    struct str * savevar;
+    int keep_c;
+
+    int a0;
+    struct str * a1;
+
+    int golab;
+    int label;
 
     if (p->left->type == c_grouping || p->left->type == c_non) {
         /* Special case for "goto" or "gopast" when used on a grouping or an
@@ -524,25 +532,27 @@ static void generate_GO(struct generator * g, struct node * p, int style) {
 #ifdef OPTIMISATION_WARNINGS
         printf("Optimising %s %s\n", style ? "goto" : "gopast", p->left->type == c_non ? "non" : "grouping");
 #endif
+        write_comment(g, p);
         generate_GO_grouping(g, p->left, style, p->left->type == c_non);
         return;
     }
 
-    int end_unreachable = false;
-    struct str * savevar = vars_newname(g);
-    int keep_c = style == 1 || repeat_restore(g, p->left);
+    end_unreachable = false;
+    savevar = vars_newname(g);
+    keep_c = style == 1 || repeat_restore(g, p->left);
 
-    int a0 = g->failure_label;
-    struct str * a1 = str_copy(g->failure_str);
+    a0 = g->failure_label;
+    a1 = str_copy(g->failure_str);
 
-    int golab = new_label(g);
+    golab = new_label(g);
     write_comment(g, p);
     w(g, "~Mtry:~N~+"
              "~Mwhile True:~N~+");
     if (keep_c) write_savecursor(g, p, savevar);
 
-    g->failure_label = new_label(g);
-    int label = g->failure_label;
+    label = new_label(g);
+    g->failure_label = label;
+    str_clear(g->failure_str);
     wsetlab_begin(g);
     generate(g, p->left);
 
@@ -595,11 +605,11 @@ static void generate_repeat_or_atleast(struct generator * g, struct node * p, st
 
     struct str * savevar = vars_newname(g);
     int keep_c = repeat_restore(g, p->left);
+    int label = new_label(g);
+    g->failure_label = label;
     writef(g, "~Mwhile True:~N~+", p);
     if (keep_c) write_savecursor(g, p, savevar);
 
-    g->failure_label = new_label(g);
-    int label = g->failure_label;
     str_clear(g->failure_str);
     wsetlab_begin(g);
     generate(g, p->left);
@@ -689,9 +699,17 @@ static void generate_hop(struct generator * g, struct node * p) {
     generate_AE(g, p->AE);
     w(g, "~N");
 
-    g->S[0] = p->mode == m_forward ? "0" : "self.limit_backward";
-
-    write_failure_if(g, "~S0 > c or c > self.limit", p);
+    g->S[1] = p->mode == m_forward ? "> self.limit" : "< self.limit_backward";
+    g->S[2] = p->mode == m_forward ? "<" : ">";
+    if (p->AE->type == c_number) {
+        // Constant distance hop.
+        //
+        // No need to check for negative hop as that's converted to false by
+        // the analyser.
+        write_failure_if(g, "c ~S1", p);
+    } else {
+        write_failure_if(g, "c ~S1 or c ~S2 self.cursor", p);
+    }
     writef(g, "~Mself.cursor = c~N", p);
 }
 
@@ -808,7 +826,7 @@ static void generate_setlimit(struct generator * g, struct node * p) {
     struct str * savevar = vars_newname(g);
     struct str * varname = vars_newname(g);
     write_comment(g, p);
-    if (p->left && p->left->type == c_tomark && !p->left->right) {
+    if (p->left && p->left->type == c_tomark) {
         /* Special case for:
          *
          *   setlimit tomark AE for C
@@ -1027,9 +1045,13 @@ static void generate_among(struct generator * g, struct node * p) {
     } else if (x->command_count > 0) {
         int i;
         for (i = 1; i <= x->command_count; i++) {
-            g->I[0] = i;
-            w(g, (i > 1 ? "~Melif" : "~Mif"));
-            w(g, " among_var == ~I0:~N~+");
+            if (i == x->command_count && x->nocommand_count == 0) {
+                w(g, "~Melse:~N~+");
+            } else {
+                g->I[0] = i;
+                w(g, (i > 1 ? "~Melif" : "~Mif"));
+                w(g, " among_var == ~I0:~N~+");
+            }
             generate(g, x->commands[i - 1]);
             w(g, "~-");
             g->unreachable = false;
@@ -1224,7 +1246,8 @@ static void generate_grouping_table(struct generator * g, struct grouping * q) {
 static void generate_groupings(struct generator * g) {
     struct grouping * q;
     for (q = g->analyser->groupings; q; q = q->next) {
-        generate_grouping_table(g, q);
+        if (q->name->used)
+            generate_grouping_table(g, q);
     }
 }
 
